(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2013-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2012, 2015)
Author:  Sarah Winkler <sarah.winkler@uibk.ac.at> (2015)
Author:  Thomas Sternagel <thomas.sternagel@uibk.ac.at> (2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Conditional_Rewriting_Impl
imports
  Conditional_Rewriting
  Trs_Impl
begin

type_synonym ('f, 'v) crules = "('f, 'v) crule list"

fun shows_conditions :: "('f :: show, 'v :: show) rule list \<Rightarrow> bool \<Rightarrow> shows"
where
  "shows_conditions [] _ = id" |
  "shows_conditions (x # xs) b = (shows (if b then '' | '' else '', '') +@+ 
    shows_rule' shows shows '' ->* '' x +@+ shows_conditions xs False)"

definition shows_crule' :: "('f \<Rightarrow> shows) \<Rightarrow> ('v \<Rightarrow> shows) \<Rightarrow> ('f, 'v) crule \<Rightarrow> shows"
where
  "shows_crule' fun var cr =
    shows_rule' fun var '' -> '' (fst cr) \<circ>
    shows_list_gen (shows_rule' fun var '' ->* '') '''' '' | '' '', '' '''' (snd cr)"

abbreviation "shows_crule \<equiv> shows_crule' shows shows"

definition
  shows_crules' :: "('f \<Rightarrow> shows) \<Rightarrow> ('v \<Rightarrow> shows) \<Rightarrow> ('f, 'v) crules \<Rightarrow> shows"
where
  "shows_crules' fun var ctrs =
    shows_list_gen (shows_crule' fun var) [] [] [CHR ''\<newline>''] [] ctrs \<circ> shows_nl"

abbreviation "shows_crules \<equiv> shows_crules' shows shows"

definition
  shows_ctrs' :: "('f \<Rightarrow> shows) \<Rightarrow> ('v \<Rightarrow> shows) \<Rightarrow> ('f, 'v) crules \<Rightarrow> shows"
where
  "shows_ctrs' fun var R = shows ''CTRS:'' \<circ> shows_nl \<circ> shows_nl \<circ> shows_crules' fun var R"

abbreviation "shows_ctrs \<equiv> shows_ctrs' shows shows"

definition check_type3 :: "('f\<Colon>show, 'v\<Colon>show) crules \<Rightarrow> shows check"
where
  "check_type3 R = do {
    check_allm (\<lambda>cr. check_subseteq (vars_term_list (snd (fst cr))) (vars_term_list (fst (fst cr)) @ vars_trs_list (snd cr))
     <+? (\<lambda>x. ''variable '' +#+ shows x
        +@+ '' occurs only in right-hand side of rule '' +#+ shows_crule cr +@+ shows_nl)) R
  } <+? (\<lambda>e. ''the CTRS is not of type 3'' +#+ shows_nl +@+ e)"

lemma check_type3 [simp]:
  "isOK (check_type3 R) = type3 (set R)"
  unfolding check_type3_def type3_def by force

definition X_impl :: "('f, 'v) crule \<Rightarrow> nat \<Rightarrow> 'v list"
 where "X_impl cr i = 
  concat (vars_term_list (fst (fst cr)) # (map (vars_term_list \<circ> snd) (take i (snd cr)) ))"

lemma X_impl [simp]:
  "set (X_impl cr i) = X_vars cr i" 
unfolding X_impl_def X_vars_def by force

definition Y_impl :: "('f, 'v) crule \<Rightarrow> nat \<Rightarrow> 'v list"
 where "Y_impl cr i = 
  vars_term_list (snd (fst cr)) @ vars_term_list (snd ((snd cr) ! i)) @ (vars_trs_list (drop (Suc i) (snd cr)))"

lemma Y_impl: 
 "set (Y_impl cr i) = Y_vars cr i" 
 unfolding Y_impl_def Y_vars_def by force

definition check_dctrs :: "('f\<Colon>show, 'v\<Colon>show) crules \<Rightarrow> shows check"
where
  "check_dctrs R = do {
    check_allm 
     (\<lambda>cr.
      check_allm
       (\<lambda>i.
        check_subseteq (vars_term_list (fst (snd cr ! i))) (X_impl cr i)
         <+? (\<lambda>x. ''variable '' +#+ shows x +@+ '' in condition '' +#+ shows_rule (snd cr ! i) +@+ '' of rule '' 
                +#+  shows_crule cr +@+ ''violates DCTRS condition'' +#+ shows_nl)
       ) [0..<length (snd cr)]
     ) R 
  } <+? (\<lambda>e. ''the CTRS is not deterministic'' +#+ shows_nl +@+ e)"

lemma check_dctrs [simp]:
  "isOK (check_dctrs R) = dctrs (set R)"
by (auto simp add: check_dctrs_def dctrs_def atLeast0LessThan)

definition check_wf_ctrs :: "('f\<Colon>show, 'v\<Colon>show) crules \<Rightarrow> shows check"
where
  "check_wf_ctrs R = do {
    check_allm (\<lambda>cr. 
      check (is_Fun (fst (fst cr)))
        (''variable left-hand side in rule '' +#+ shows_crule cr  +@+ shows_nl)
    ) R;
    check_dctrs R;
    check_type3 R
  } <+? (\<lambda>e. ''the CTRS is not well-formed'' +#+ shows_nl +@+ e)"

lemma check_wf_ctrs [simp]:
  "isOK (check_wf_ctrs R) = wf_ctrs (set R)"
  unfolding check_wf_ctrs_def wf_ctrs_def
  by force+

definition funs_crule_list :: "('f, 'v) crule \<Rightarrow> 'f list"
where
  "funs_crule_list r = add_funs_rule (fst r) (funs_trs_list (snd r))"

lemma funs_crule_list_sound [simp]:
  "set (funs_crule_list r) = funs_crule r"
  by (simp add: funs_crule_list_def funs_crule_def)

definition funs_ctrs_list :: "('f, 'v) crule list \<Rightarrow> 'f list"
where
  "funs_ctrs_list trs = concat (map funs_crule_list trs)"

lemma funs_ctrs_list_sound [simp]:
  "set (funs_ctrs_list r) = funs_ctrs (set r)"
  by (simp add: funs_ctrs_list_def funs_ctrs_def)

abbreviation "terms_of_rules rs \<equiv> map fst rs @ map snd rs"
abbreviation "terms_of_crule \<rho> \<equiv> (clhs \<rho> # crhs \<rho> # terms_of_rules (snd \<rho>))"

definition "match_rules rs\<^sub>1 rs\<^sub>2 = do {
  ts \<leftarrow> zip_option (terms_of_rules rs\<^sub>2) (terms_of_rules rs\<^sub>1);
  match_list Var ts
}"

definition "match_crule \<rho>\<^sub>1 \<rho>\<^sub>2 = do {
  ts \<leftarrow> zip_option (terms_of_crule \<rho>\<^sub>2) (terms_of_crule \<rho>\<^sub>1);
  match_list Var ts
}"

abbreviation "term_of_rules rs \<equiv> Fun (SOME f. True) (terms_of_rules rs)"
abbreviation "term_of_crule \<rho> \<equiv> Fun (SOME f. True) (clhs \<rho> # crhs \<rho> # terms_of_rules (snd \<rho>))"

lemma match_crule_alt_def:
  "match_crule \<rho>\<^sub>1 \<rho>\<^sub>2 = match (term_of_crule \<rho>\<^sub>1) (term_of_crule \<rho>\<^sub>2)"
by (cases "zip_option (terms_of_crule \<rho>\<^sub>1) (terms_of_crule \<rho>\<^sub>2)")
   (auto simp: match_crule_def match_def match_list_def decompose_def split: option.splits)

lemma match_rules_alt_def:
  "match_rules rs\<^sub>1 rs\<^sub>2 = match (term_of_rules rs\<^sub>1) (term_of_rules rs\<^sub>2)"
by (cases "zip_option (terms_of_rules rs\<^sub>1) (terms_of_rules rs\<^sub>2)")
   (auto simp: match_rules_def match_def match_list_def decompose_def split: option.splits)

end
