(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2015)
Author:  Thomas Sternagel <thomas.sternagel@uibk.ac.at> (2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Level_Confluence_Impl
imports
  "../Rewriting/Tcap_Impl" (*FIXME: check why order matters for adhoc overloading*)
  Level_Confluence
  Conditional_Rewriting_Impl
begin

(*TODO: move*)
abbreviation check_exists :: "('a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> ('a list) check"
where
  "check_exists f xs \<equiv> check_exm (\<lambda>x. if f x then succeed else error [x]) xs concat"

(*TODO: move to AFP*)
fun choice :: "('e + 'a) list \<Rightarrow> 'e list + 'a"
where
  "choice [] = error []" |
  "choice (x # xs) = (try x catch (\<lambda>e. choice xs <+? Cons e))"

lemma isOK_choice [simp]:
  "isOK (choice []) \<longleftrightarrow> False"
  "isOK (choice (x # xs)) \<longleftrightarrow> isOK x \<or> isOK (choice xs)"
by (auto simp: isOK_def split: sum.splits)

declare choice.simps [simp del]

(*TODO: move*)
fun funposs_list :: "('f, 'v) term \<Rightarrow> pos list"
where
  "funposs_list (Var x) = []" |
  "funposs_list (Fun f ss) = (\<epsilon> # concat (map (\<lambda> (i, ps).
    map (op <# i) ps) (zip [0 ..< length ss] (map funposs_list ss))))"

(*TODO: move*)
lemma UNION_set_zip:
  "(\<Union>x \<in> set (zip [0..<length xs] (map f xs)). g x) = (\<Union>i < length xs. g (i, f (xs ! i)))"
by (auto simp: set_conv_nth)

(*TODO: move*)
lemma set_funposs_list [simp]:
  "set (funposs_list t) = funposs t"
by (induct t; auto simp: UNION_set_zip)

definition check_properly_oriented :: "('f\<Colon>show, 'v\<Colon>show) crules \<Rightarrow> shows check"
where
  "check_properly_oriented R =
    check_dctrs (filter (\<lambda>((l, r), cs). \<not> vars_term r \<subseteq> vars_term l) R)
    <+? (\<lambda>e. ''the CTRS is not properly oriented'' +#+ shows_nl +@+ e)"
(* export_code check_properly_oriented in Haskell *)

lemma check_properly_oriented [simp]:
  "isOK (check_properly_oriented R) = properly_oriented (set R)"
  by (simp add: check_properly_oriented_def properly_oriented_def X_vars_def dctrs_def) (fastforce)

(* TODO: move *)
definition check_linear_term :: "('f\<Colon>show, 'v\<Colon>show) term \<Rightarrow> shows check"
where
  "check_linear_term s = check (linear_term s)
    (shows ''the term '' \<circ> shows_term s \<circ> shows '' is not linear'' \<circ> shows_nl)"

lemma check_linear_term [simp]:
  "isOK (check_linear_term s) = linear_term s"
by (simp add: check_linear_term_def)

definition check_constructor_term ::
  "('f\<Colon>show, 'v\<Colon>show) term \<Rightarrow> ('f\<Colon>show, 'v\<Colon>show) crules \<Rightarrow> shows check"
where
  "check_constructor_term s R = check
    (funas_term s \<subseteq> funas_ctrs (set R) - set (defined_list (map fst R)))
    (shows ''the term '' \<circ> shows_term s \<circ> shows '' is not a constructor term'' \<circ> shows_nl)"

lemma check_constructor_term [simp]:
  "isOK (check_constructor_term s R) = (funas_term s \<subseteq> funas_ctrs (set R) - { f. defined (Ru (set R)) f })"
by (simp add: check_constructor_term_def Ru_def)

definition check_ground_term :: "('f\<Colon>show, 'v\<Colon>show) term \<Rightarrow> shows check"
where
  "check_ground_term s = check (ground s)
    (shows ''the term '' \<circ> shows_term s \<circ> shows '' is not a ground term'' \<circ> shows_nl)"

lemma check_ground_term [simp]:
  "isOK (check_ground_term s) \<longleftrightarrow> ground s"
by (simp add: check_ground_term_def)

definition check_normal_form :: 
  "('f\<Colon>{key,show}, 'v\<Colon>show) term \<Rightarrow> ('f, 'v) crules \<Rightarrow> shows check"
where
  "check_normal_form s R = check
    (is_NF_trs (map fst R) s)
    (shows ''the term '' \<circ> shows_term s \<circ> shows '' is not an Ru normal form'' \<circ> shows_nl)"

lemma check_normal_form [simp]:
  "isOK (check_normal_form s R) \<longleftrightarrow> s \<in> NF (rstep (Ru (set R)))"
by (simp add: check_normal_form_def Ru_def)

definition check_right_stable :: "('f\<Colon>{key,show}, 'v\<Colon>show) crules \<Rightarrow> shows check"
where
  "check_right_stable R = check_allm (\<lambda>r.
    check_allm (\<lambda>i. do {
      let t\<^sub>i = snd (snd r ! i);
      check_disjoint
        (vars_term_list (fst (fst r)) @
        (concat (map (\<lambda>(s, t). vars_term_list s) (take (Suc i) (snd r)))) @
        (concat (map (\<lambda>(s, t). vars_term_list t) (take i (snd r)))))
        (vars_term_list t\<^sub>i)
      <+? (\<lambda>x. shows ''variable '' \<circ> shows x \<circ> shows '' in rhs of condition '' \<circ>
               shows i \<circ> shows '' is not fresh'' \<circ> shows_nl);
      choice [(check_linear_term t\<^sub>i \<guillemotright> check_constructor_term t\<^sub>i R),
        (check_ground_term t\<^sub>i \<guillemotright> check_normal_form t\<^sub>i R)] <+? shows_sep id shows_nl
    }
    ) [0..< length (snd r)]
  ) R
  <+? (\<lambda>e. shows ''the CTRS is not right stable'' \<circ> shows_nl \<circ> e)"

lemma check_right_stable [simp]:
  "isOK (check_right_stable R) = right_stable (set R)"
unfolding check_right_stable_def
using [[simproc del: list_to_set_comprehension]]
by (simp add: right_stable_def Let_def atLeast0LessThan case_prod_beta) (blast)

lemma crule_cases:
  assumes "\<And>l r cs. \<rho> = ((l, r), cs) \<Longrightarrow> P"
  shows "P"
using assms by (cases \<rho>; auto)

lemma rules_enc_aux:
  fixes xs ys :: "('f, 'v :: infinite) rules"
  assumes "map (op \<bullet> \<pi> \<circ> fst) xs @ map (op \<bullet> \<pi> \<circ> snd) xs = map fst ys @ map snd ys"
  shows "map (op \<bullet> \<pi>) xs = ys"
proof -
  have "length xs = length ys"
  proof (rule ccontr)
    assume "length xs \<noteq> length ys"
    moreover
    have "length (map (op \<bullet> \<pi> \<circ> fst) xs @ map (op \<bullet> \<pi> \<circ> snd) xs) = length (map fst ys @ map snd ys)"
      using assms by simp
    ultimately show False by simp
  qed
  then show ?thesis
    using assms by (induct xs ys rule: list_induct2; auto simp: eqvt)
qed

lemma perm_term_of_crule_eq_conv:
  "\<pi> \<bullet> term_of_crule \<rho>\<^sub>1 = term_of_crule \<rho>\<^sub>2 \<longleftrightarrow> \<pi> \<bullet> \<rho>\<^sub>1 = \<rho>\<^sub>2"
by (cases \<rho>\<^sub>1 rule: crule_cases; cases \<rho>\<^sub>2 rule: crule_cases; auto simp: eqvt rules_enc_aux)

lemma perm_term_of_rules_eq_conv:
  "\<pi> \<bullet> term_of_rules rs\<^sub>1 = term_of_rules rs\<^sub>2 \<longleftrightarrow> \<pi> \<bullet> rs\<^sub>1 = rs\<^sub>2"
by (auto simp: eqvt rules_enc_aux)

lemma match_crule_imp_variants:
  fixes \<rho>\<^sub>1 \<rho>\<^sub>2 :: "('f, 'v :: infinite) crule"
  assumes "match_crule \<rho>\<^sub>2 \<rho>\<^sub>1 = Some \<sigma>\<^sub>1" and "match_crule \<rho>\<^sub>1 \<rho>\<^sub>2 = Some \<sigma>\<^sub>2" 
  shows "\<exists>\<pi>. \<pi> \<bullet> \<rho>\<^sub>1 = \<rho>\<^sub>2"
proof -
  have "term_of_crule \<rho>\<^sub>2 \<cdot> \<sigma>\<^sub>2 = term_of_crule \<rho>\<^sub>1"
    and "term_of_crule \<rho>\<^sub>1 \<cdot> \<sigma>\<^sub>1 = term_of_crule \<rho>\<^sub>2"
    using assms by (auto simp: match_crule_alt_def dest: match_sound)
  then obtain \<pi> where "\<pi> \<bullet> \<rho>\<^sub>2 = \<rho>\<^sub>1"
    using term_variants_iff [of "term_of_crule \<rho>\<^sub>2" "term_of_crule \<rho>\<^sub>1"]
    unfolding perm_term_of_crule_eq_conv by auto
  then have "-\<pi> \<bullet> \<rho>\<^sub>1 = \<rho>\<^sub>2" by auto
  then show ?thesis ..
qed

lemma match_rules_imp_variants:
  fixes rs\<^sub>1 rs\<^sub>2 :: "('f, 'v :: infinite) rules"
  assumes "match_rules rs\<^sub>2 rs\<^sub>1 = Some \<sigma>\<^sub>1" and "match_rules rs\<^sub>1 rs\<^sub>2 = Some \<sigma>\<^sub>2"
  shows "\<exists>\<pi>::'v perm. \<pi> \<bullet> rs\<^sub>1 = rs\<^sub>2"
proof -
  have "term_of_rules rs\<^sub>2 \<cdot> \<sigma>\<^sub>2 = term_of_rules rs\<^sub>1"
    and "term_of_rules rs\<^sub>1 \<cdot> \<sigma>\<^sub>1 = term_of_rules rs\<^sub>2"
    using assms by (auto simp: match_rules_alt_def dest: match_sound)
  then obtain \<pi> :: "'v perm" where "\<pi> \<bullet> rs\<^sub>2 = rs\<^sub>1"
    using term_variants_iff [of "term_of_rules rs\<^sub>2" "term_of_rules rs\<^sub>1"]
    unfolding perm_term_of_rules_eq_conv by auto
  then have "-\<pi> \<bullet> rs\<^sub>1 = rs\<^sub>2" by (auto simp del: rules_pt.permute_list_def)
  then show ?thesis ..
qed

text \<open>
  Operations required to support a check function for almost-orthogonality (modulo infeasibility)
\<close>
locale almost_orthogonal_ops =
  fixes xvar yvar :: "'v::{show, infinite} \<Rightarrow> 'v"
    and check_infeasible ::
      "('f::show, 'v) crules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> shows check"
begin

definition "shows_overlap \<rho>\<^sub>1 \<rho>\<^sub>2 p =
  shows ''the overlap of rule '' \<circ> shows_crule \<rho>\<^sub>1 \<circ> shows '' with rule '' \<circ> shows_crule \<rho>\<^sub>2 \<circ>
  shows '' at position '' \<circ> shows p"

definition
  check_overlap :: "('f, 'v) crules \<Rightarrow> ('f, 'v) crule \<Rightarrow> ('f, 'v) crule \<Rightarrow> pos \<Rightarrow> shows check"
where
  "check_overlap R \<rho>\<^sub>1 \<rho>\<^sub>2 p =
    (case mgu_var_disjoint_generic xvar yvar (clhs \<rho>\<^sub>1 |_ p) (clhs \<rho>\<^sub>2) of
      None \<Rightarrow> succeed
    | Some (\<sigma>\<^sub>1, \<sigma>\<^sub>2) \<Rightarrow>
      choice [
        check (p = \<epsilon> \<and> crhs \<rho>\<^sub>1 \<cdot> \<sigma>\<^sub>1 = crhs \<rho>\<^sub>2 \<cdot> \<sigma>\<^sub>2) (shows ''is not a trivial root-overlap''),
        check (p = \<epsilon> \<and> match_crule \<rho>\<^sub>1 \<rho>\<^sub>2 \<noteq> None \<and> match_crule \<rho>\<^sub>2 \<rho>\<^sub>1 \<noteq> None)
          (shows ''is not a root-overlap of variants of the same rule''),
        check_infeasible R (subst_list \<sigma>\<^sub>1 (snd \<rho>\<^sub>1)) (subst_list \<sigma>\<^sub>2 (snd \<rho>\<^sub>2))
          <+? (\<lambda>e. shows ''could not be shown to be infeasible'' \<circ> shows_nl \<circ> e)]
      <+? shows_sep id shows_nl)
    <+? (\<lambda>e. shows_overlap \<rho>\<^sub>1 \<rho>\<^sub>2 p \<circ> shows_nl \<circ> e)"

definition
  "check_ao R = do {
    check_left_linear_trs (map fst R);
    check_allm (\<lambda>\<rho>\<^sub>1. let l\<^sub>1 = clhs \<rho>\<^sub>1 in check_allm (\<lambda>\<rho>\<^sub>2. check_allm (\<lambda>p.
      check_overlap R \<rho>\<^sub>1 \<rho>\<^sub>2 p) (funposs_list l\<^sub>1)) R) R
  }"

end

lemmas [code] =
  almost_orthogonal_ops.check_overlap_def almost_orthogonal_ops.check_ao_def

lemma cstep_n_permute_iff:
  "(\<pi> \<bullet> s, \<pi> \<bullet> t) \<in> cstep_n R n \<longleftrightarrow> (s, t) \<in> cstep_n R n"
by (auto simp: cstep_n_rstep_trs_n_conv split: prod.splits)

lemma subst_compose_o_assoc:
  "(\<sigma> \<circ>\<^sub>s \<tau>) \<circ> f = (\<sigma> \<circ> f) \<circ>\<^sub>s \<tau>"
by (rule ext) (simp add: subst_compose)

lemma perm_csteps_n_perm:
  assumes "(\<pi> \<bullet> s, t) \<in> (cstep_n R n)\<^sup>*"
  shows "\<exists>u. t = \<pi> \<bullet> u"
  using assms by (metis term_pt.permute_minus_cancel(1))

lemma subst_list_subst_compose:
  "subst_list (\<sigma> \<circ>\<^sub>s \<tau>) xs = subst_list \<tau> (subst_list \<sigma> xs)"
by (simp add: subst_list_def)

lemma subst_list_Rep_perm:
  "subst_list (\<sigma> \<circ> Rep_perm \<pi>) xs = subst_list \<sigma> (\<pi> \<bullet> xs)"
by (auto simp: subst_list_def permute_term_subst_apply_term [symmetric] eqvt)

locale almost_orthogonal_spec = almost_orthogonal_ops +
  assumes infeasible: "isOK (check_infeasible R cs\<^sub>1 cs\<^sub>2) \<Longrightarrow>
    (\<forall>m n. comm ((cstep_n (set R) m)\<^sup>*) ((cstep_n (set R) n)\<^sup>*) \<longrightarrow>
    \<not> (\<exists>\<sigma>. conds_sat (set R) m cs\<^sub>1 \<sigma> \<and> conds_sat (set R) n cs\<^sub>2 \<sigma>))"
    and ren: "inj xvar" "inj yvar" "range xvar \<inter> range yvar = {}"
begin

lemma isOK_check_ao:
  assumes "isOK (check_ao R)"
  shows "almost_orthogonal (set R)"
proof (unfold almost_orthogonal_def, intro conjI allI impI)
  show "left_linear_trs (Ru (set R))"
    using assms by (simp add: check_ao_def Ru_def)
next
  fix \<rho>\<^sub>1' \<rho>\<^sub>2' p
  assume "overlap (set R) \<rho>\<^sub>1' \<rho>\<^sub>2' p"
  then obtain \<pi>\<^sub>1 and \<pi>\<^sub>2 and \<mu>
    where rules': "\<pi>\<^sub>1 \<bullet> \<rho>\<^sub>1' \<in> set R" "\<pi>\<^sub>2 \<bullet> \<rho>\<^sub>2' \<in> set R"
    and "vars_crule \<rho>\<^sub>1' \<inter> vars_crule \<rho>\<^sub>2' = {}"
    and p': "p \<in> funposs (clhs \<rho>\<^sub>1')"
    and mgu': "mgu (clhs \<rho>\<^sub>1' |_ p) (clhs \<rho>\<^sub>2') = Some \<mu>"
    by (auto simp: overlap_def)
  def \<rho>\<^sub>1 \<equiv> "\<pi>\<^sub>1 \<bullet> \<rho>\<^sub>1'" and \<rho>\<^sub>2 \<equiv> "\<pi>\<^sub>2 \<bullet> \<rho>\<^sub>2'"
  have p: "p \<in> funposs (clhs \<rho>\<^sub>1)"
    and "\<rho>\<^sub>1 \<in> set R" and "\<rho>\<^sub>2 \<in> set R"
    using p' and rules' apply (auto simp: \<rho>\<^sub>1_def \<rho>\<^sub>2_def eqvt)
    by (metis crule_pt.fst_eqvt funposs_perm_simp rule_pt.fst_eqvt)
  with assms have *: "isOK (check_overlap R \<rho>\<^sub>1 \<rho>\<^sub>2 p)" by (auto simp: check_ao_def)

  let ?thesis = "let \<mu> = the (mgu (clhs \<rho>\<^sub>1' |_ p) (clhs \<rho>\<^sub>2')) in
     p = \<epsilon> \<and> crhs \<rho>\<^sub>1' \<cdot> \<mu> = crhs \<rho>\<^sub>2' \<cdot> \<mu> \<or>
     p = \<epsilon> \<and> (\<exists>p. p \<bullet> \<rho>\<^sub>1' = \<rho>\<^sub>2') \<or>
     (\<forall>m n. comm ((cstep_n (set R) m)\<^sup>*) ((cstep_n (set R) n)\<^sup>*) \<longrightarrow>
       \<not> (\<exists>\<sigma>. conds_sat (set R) m (subst_list \<mu> (snd \<rho>\<^sub>1')) \<sigma> \<and>
              conds_sat (set R) n (subst_list \<mu> (snd \<rho>\<^sub>2')) \<sigma>))"

  have "(clhs \<rho>\<^sub>1' |_ p) \<cdot> \<mu> = clhs \<rho>\<^sub>2' \<cdot> \<mu>" using mgu' [THEN mgu_sound] by (auto simp: is_imgu_def)
  then have "(clhs \<rho>\<^sub>1' |_ p) \<cdot> (\<mu> \<circ> Rep_perm (-\<pi>\<^sub>1) \<circ> Rep_perm \<pi>\<^sub>1) =
    clhs \<rho>\<^sub>2' \<cdot> (\<mu> \<circ> Rep_perm (-\<pi>\<^sub>2) \<circ> Rep_perm \<pi>\<^sub>2)"
    by (simp add: o_assoc [symmetric] Rep_perm_add [symmetric] Rep_perm_0)
  then have "clhs \<rho>\<^sub>1 |_ p \<cdot> (\<mu> \<circ> Rep_perm (-\<pi>\<^sub>1)) = clhs \<rho>\<^sub>2 \<cdot> (\<mu> \<circ> Rep_perm (-\<pi>\<^sub>2))"
    using p [THEN funposs_imp_poss]
    by (auto simp: \<rho>\<^sub>1_def \<rho>\<^sub>2_def permute_term_subst_apply_term [symmetric] eqvt)
  from mgu_var_disjoint_generic_complete [OF ren this] obtain \<mu>\<^sub>1 \<mu>\<^sub>2 \<delta>
    where "mgu_var_disjoint_generic xvar yvar (clhs \<rho>\<^sub>1 |_ p) (clhs \<rho>\<^sub>2) = Some (\<mu>\<^sub>1, \<mu>\<^sub>2)"
    and **: "\<mu> \<circ> Rep_perm (- \<pi>\<^sub>1) = \<mu>\<^sub>1 \<circ>\<^sub>s \<delta>" "\<mu> \<circ> Rep_perm (- \<pi>\<^sub>2) = \<mu>\<^sub>2 \<circ>\<^sub>s \<delta>" by blast
  with * have "p = \<epsilon> \<and> crhs \<rho>\<^sub>1 \<cdot> \<mu>\<^sub>1 = crhs \<rho>\<^sub>2 \<cdot> \<mu>\<^sub>2 \<or>
    p = \<epsilon> \<and> match_crule \<rho>\<^sub>1 \<rho>\<^sub>2 \<noteq> None \<and> match_crule \<rho>\<^sub>2 \<rho>\<^sub>1 \<noteq> None \<or>
    isOK (check_infeasible R (subst_list \<mu>\<^sub>1 (snd \<rho>\<^sub>1)) (subst_list \<mu>\<^sub>2 (snd \<rho>\<^sub>2)))"
    by (auto simp: check_overlap_def)
  moreover
  { assume [simp]: "p = \<epsilon>" and "crhs \<rho>\<^sub>1 \<cdot> \<mu>\<^sub>1 = crhs \<rho>\<^sub>2 \<cdot> \<mu>\<^sub>2"
    then have "crhs \<rho>\<^sub>1 \<cdot> (\<mu>\<^sub>1 \<circ>\<^sub>s \<delta>) = crhs \<rho>\<^sub>2 \<cdot> (\<mu>\<^sub>2 \<circ>\<^sub>s \<delta>)" by simp
    then have "crhs \<rho>\<^sub>1' \<cdot> \<mu> = crhs \<rho>\<^sub>2' \<cdot> \<mu>"
      unfolding ** [symmetric]
      by (simp add: \<rho>\<^sub>1_def \<rho>\<^sub>2_def permute_term_subst_apply_term [symmetric] eqvt)
    with mgu' have ?thesis by (auto) }
  moreover
  { assume "p = \<epsilon>" and "match_crule \<rho>\<^sub>1 \<rho>\<^sub>2 \<noteq> None" and "match_crule \<rho>\<^sub>2 \<rho>\<^sub>1 \<noteq> None"
    with match_crule_imp_variants [of \<rho>\<^sub>2 \<rho>\<^sub>1] obtain \<pi> where "\<pi> \<bullet> \<rho>\<^sub>1 = \<rho>\<^sub>2" by blast
    then have "(-\<pi>\<^sub>2 + \<pi> + \<pi>\<^sub>1) \<bullet> \<rho>\<^sub>1' = \<rho>\<^sub>2'" by (simp add: \<rho>\<^sub>1_def \<rho>\<^sub>2_def)
    with \<open>p = \<epsilon>\<close> have ?thesis unfolding Let_def by blast }
  moreover
  { assume *: "isOK (check_infeasible R (subst_list \<mu>\<^sub>1 (snd \<rho>\<^sub>1)) (subst_list \<mu>\<^sub>2 (snd \<rho>\<^sub>2)))"
    { fix m n assume comm: "comm ((cstep_n (set R) m)\<^sup>*) ((cstep_n (set R) n)\<^sup>*)"
      have "\<not> (\<exists>\<sigma>. conds_sat (set R) m (subst_list \<mu> (snd \<rho>\<^sub>1')) \<sigma> \<and>
        conds_sat (set R) n (subst_list \<mu> (snd \<rho>\<^sub>2')) \<sigma>)"
      proof (intro notI, elim exE conjE)
        fix \<sigma>
        assume "conds_sat (set R) m (subst_list \<mu> (snd \<rho>\<^sub>1')) \<sigma>"
          and "conds_sat (set R) n (subst_list \<mu> (snd \<rho>\<^sub>2')) \<sigma>"
        then have "conds_sat (set R) m (subst_list (\<mu> \<circ> Rep_perm (-\<pi>\<^sub>1)) (snd \<rho>\<^sub>1)) \<sigma>"
          and "conds_sat (set R) n (subst_list (\<mu> \<circ> Rep_perm (-\<pi>\<^sub>2)) (snd \<rho>\<^sub>2)) \<sigma>"
          by (auto simp: \<rho>\<^sub>1_def \<rho>\<^sub>2_def subst_list_subst_compose subst_list_Rep_perm map_idI eqvt [symmetric])
        then have "\<exists>\<sigma>. conds_sat (set R) m (subst_list \<mu>\<^sub>1 (snd \<rho>\<^sub>1)) \<sigma> \<and>
          conds_sat (set R) n (subst_list \<mu>\<^sub>2 (snd \<rho>\<^sub>2)) \<sigma>"
          by (auto simp add: ** subst_list_subst_compose conds_sat_subst_list)
        with infeasible [OF *, THEN spec, THEN spec, THEN mp, OF comm] show False by blast
      qed }
    then have ?thesis using mgu' by auto }
  ultimately show ?thesis by blast
qed

end

interpretation wo_infeasibility!: almost_orthogonal_spec x_var y_var
  "\<lambda>R cs\<^sub>1 cs\<^sub>2. error (shows_string ''infeasibility check not supported'')"
by (unfold_locales; auto simp: inj_on_def)

definition check_almost_orthogonal :: "('f::show, string) crules \<Rightarrow> shows check"
where
  "check_almost_orthogonal R = wo_infeasibility.check_ao R"

lemma check_almost_orthogonal [simp]:
  assumes "isOK (check_almost_orthogonal R)"
  shows "almost_orthogonal (set R)"
using assms by (intro wo_infeasibility.isOK_check_ao) (simp add: check_almost_orthogonal_def)

definition check_level_confluence :: "('f::{key,show}, string) crules \<Rightarrow> shows check"
where
  "check_level_confluence R = do {
    check_allm (\<lambda>r. 
      check (is_Fun (fst (fst r)))
      (shows ''variable left-hand side in rule '' \<circ> shows_crule r  \<circ> shows_nl)
    ) R;
    check_type3 R;
    check_properly_oriented R;
    check_right_stable R;
    check_almost_orthogonal R
  }"

lemma check_level_confluence [simp]:
  assumes "isOK (check_level_confluence R)"
  shows "level_confluent (set R)"
using assms by (intro level_confluence) (auto simp: check_level_confluence_def split: prod.splits)

(*TODO: move*)
datatype non_reachability_proof =
  Not_Reachable_Tcap

datatype ('f, 'v) infeasibility_proof =
  Compound_Conditions_Not_Reachable 'f "non_reachability_proof"

fun check_not_reachable
where
  "check_not_reachable R s t Not_Reachable_Tcap =
    check (\<not> Ground_Context.match (tcapI R s) t)
      (shows ''could not show non-reachability via tcap'')"

lemma check_not_reachable [simp]:
  assumes "isOK (check_not_reachable R s t p)"
  shows "\<not> (\<exists>\<sigma> \<tau>. (s \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep (set R))\<^sup>*)"
using assms and match_tcap_sound [of s _ t _ "set R"] by (cases p) auto

fun check_infeasible'
where
  "check_infeasible' R cs (Compound_Conditions_Not_Reachable f p) =
    check_not_reachable (map fst R) (Fun f (map fst cs)) (Fun f (map snd cs)) p"

definition
  check_infeasible ::
    "(('f, 'v) rules \<times> ('f, 'v) infeasibility_proof) list \<Rightarrow>
      ('f, 'v) crules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> shows check"
where
  "check_infeasible css R cs\<^sub>1 cs\<^sub>2 = check_exm (\<lambda>(cs, p). do {
    let cs' = cs\<^sub>1 @ cs\<^sub>2;
    check (match_rules cs cs' \<noteq> None \<and> match_rules cs' cs \<noteq> None) id;
    check_infeasible' R cs p
  }) css (shows_sep id shows_nl)"

lemma check_infeasible:
  fixes R :: "('f, 'v::infinite) crules"
  assumes "isOK (check_infeasible css R cs\<^sub>1 cs\<^sub>2)"
  shows "\<not> (\<exists>\<sigma> n. conds_sat (set R) n (cs\<^sub>1 @ cs\<^sub>2) \<sigma>)"
proof -
  let ?cs' = "cs\<^sub>1 @ cs\<^sub>2"
  obtain \<pi> :: "'v perm" and cs and p where "(cs, p) \<in> set css"
    and "\<pi> \<bullet> ?cs' = cs" and ok: "isOK (check_infeasible' R cs p)"
  using assms and match_rules_imp_variants [of _ ?cs']
    by (fastforce simp: check_infeasible_def Ru_def)
  then have *: "\<pi> \<bullet> map fst ?cs' = map fst cs" "\<pi> \<bullet> map snd ?cs' = map snd cs" by (auto simp: eqvt)
  show ?thesis
  proof (cases p)
    case (Compound_Conditions_Not_Reachable f q)
    note infeasible = this
    let ?s = "Fun f (map fst cs)" and ?s' = "Fun f (map fst ?cs')"
    let ?t = "Fun f (map snd cs)" and ?t' = "Fun f (map snd ?cs')"
    have **: "\<not> (\<exists>\<sigma>. (?s \<cdot> \<sigma>, ?t \<cdot> \<sigma>) \<in> (rstep (Ru (set R)))\<^sup>*)"
      using infeasible and ok by (auto simp: Ru_def dest: check_not_reachable)
    { fix \<sigma> :: "('f, 'v) subst" and n assume "conds_sat (set R) n ?cs' \<sigma>"
      then have "\<forall>(s, t) \<in> set ?cs'. (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (rstep (Ru (set R)))\<^sup>*"
        using rtrancl_mono [OF cstep_imp_Ru_step, of "set R"]
        by (auto simp: conds_sat_iff split: prod.splits dest!: csteps_n_imp_csteps)
      then have "(?s' \<cdot> \<sigma>, ?t' \<cdot> \<sigma>) \<in> (rstep (Ru (set R)))\<^sup>*"
        apply (unfold subst_apply_term.simps, intro args_steps_imp_steps)
        apply (auto dest!: nth_mem split: prod.splits simp: nth_append)
        apply (metis Un_iff fst_conv snd_conv surj_pair)
        by (metis in_set_conv_nth length_append nth_append prod.collapse set_append)
      then have "(\<pi> \<bullet> (?s' \<cdot> \<sigma>), \<pi> \<bullet> (?t' \<cdot> \<sigma>)) \<in> (rstep (Ru (set R)))\<^sup>*"
        using rstep_rtrancl_idemp by blast
      with ** have False
        by (auto simp only: term_apply_subst_eqvt permute_term.simps *
                            terms_pt.permute_list_def [symmetric]) }
    then show ?thesis by blast
  qed
qed

interpretation mod_infeasibility!: almost_orthogonal_spec x_var y_var
  "check_infeasible css" for css
by (unfold_locales; force dest: infeasible_sufficient check_infeasible simp: inj_on_def)

definition
  check_almost_orthogonal_modulo_infeasibility ::
    "_ \<Rightarrow> ('f::show, string) crules \<Rightarrow> shows check"
where
  "check_almost_orthogonal_modulo_infeasibility css R = mod_infeasibility.check_ao css R"

lemma check_almost_orthogonal_moduleo_infeasibility [simp]:
  assumes "isOK (check_almost_orthogonal_modulo_infeasibility css R)"
  shows "almost_orthogonal (set R)"
using assms
by (intro mod_infeasibility.isOK_check_ao)
   (simp add: check_almost_orthogonal_modulo_infeasibility_def)

definition
  check_level_confluence_modulo_infeasibility ::
    "_ \<Rightarrow> ('f::{key,show}, string) crules \<Rightarrow> shows check"
where
  "check_level_confluence_modulo_infeasibility css R = do {
    check_allm (\<lambda>r. 
      check (is_Fun (fst (fst r)))
      (shows ''variable left-hand side in rule '' \<circ> shows_crule r  \<circ> shows_nl)
    ) R;
    check_type3 R;
    check_properly_oriented R;
    check_right_stable R;
    check_almost_orthogonal_modulo_infeasibility css R
  }"

lemma check_level_confluence_moduleo_infeasibility [simp]:
  assumes "isOK (check_level_confluence_modulo_infeasibility css R)"
  shows "level_confluent (set R)"
using assms
by (intro level_confluence)
   (auto simp: check_level_confluence_modulo_infeasibility_def split: prod.splits)

end
